"""A helper tool to convert CSV files to ARFF files."""


import sys
sys.path.append("..")

from DataAccess import Attr, AttrSet
from DataAccess import make_arff_header




infile = "/home/szymon/dmining/data/DrCiechanowicz/dSS/dSS_train.csv"
missing_strs = [""]  # strings to be treated as missing values




class attr_string_stats(object):
    """Statistics for guessing attribute's type"""
    maxcategs = 50  # max. # of categories
    def __init__(self, name, missing_strs = [""]):
        self.name = name
        self.domain = set()
        self.n_numeric = 0
        self.N = 0
        self.missing_strs = missing_strs

    def update(self, strval):
        if strval in self.missing_strs:
            return
        self.N += 1
        if len(self.domain) < self.maxcategs:
            self.domain.add(strval)
        try:
            fv = float(strval)
            self.n_numeric += 1
        except:
            pass

    def get_attr(self):
        """Return the guessed attribute."""
        # first guess attribute's type
        # if all numbers but small consecutive ints, then categ
        intdomain = set()
        for v in self.domain:
            try:
                iv = int(v)
                if iv >= 0:
                    intdomain.add(iv)
            except:
                pass
        intdomain = list(intdomain)
        intdomain.sort()
        
        if (len(intdomain) > 0 and len(intdomain) >= len(self.domain) - 2
            and intdomain[0] < 2 and intdomain[-1] <= len(intdomain) + 5): # allow for 2 special and 5 missing from sequence values
            t = "CATEG"
        elif len(self.domain) < 20 and self.n_numeric <= 0.7 * self.N:
            t = "CATEG"
        elif self.n_numeric > 0.7 * self.N:
            t = "CONTINUOUS"
        else:
            t = "STRING"

        # now create the attribute
        if t == "CATEG":
            dom = list(self.domain)
            dom.sort()
            attr = Attr(self.name, t, dom)
        else:
            attr = Attr(self.name, t)
        return attr
    

    def __str__(self):
        ret = ""
        ret += "domain size %d (%s)\n" % (len(self.domain), self.domain)
        ret += "%d numeric values\n" % self.n_numeric
        return ret



f = file(infile, "rU")
header_line = f.readline()
attrnames = [name.strip().strip("'\"") for name in header_line.split(",")]

attr_stats = [attr_string_stats(aname, missing_strs) for aname in attrnames]
for line in f:
    vals = [name.strip().strip("'\"") for name in line.split(",")]
    for v, s in zip(vals, attr_stats):
        s.update(v)

#for s in attr_stats:
#    print s.get_attr()

# create the header
aset = AttrSet("noname", [s.get_attr() for s in attr_stats])
print make_arff_header(aset),

# print data
f.seek(0)
f.readline()
for line in f:
    print line,
