import shlex
import numpy

from DataAccess import Attr
from BayesNet import BayesNet
from NoisyOR import NoisyOR


class BNReadError(RuntimeError):
    pass

def check_token(string, tok, lex):
    """Assert that tok==string."""
    if tok != string:
	msg = lex.error_leader() + "'" + string + "' expected, got '" + tok + "' instead"
	raise RuntimeError, msg
    
def match_token(string, lex):
    """Get given string from the lexer."""
    tok = lex.get_token()
    check_token(string, tok, lex)

def parse_list(lex):
    lst = []
    tok = lex.get_token()
    while tok != ")" and tok != "":
        lex.push_token(tok)
        val = parse_value(lex)
        lst.append(val)
	tok = lex.get_token()
    check_token(")", tok, lex)
    #lex.push_token(tok)
    #match_token(")", lex)
    return lst
    
def parse_value(lex):
    tok = lex.get_token()
    #print tok, len(tok), tok == "\""
    if len(tok) == 0:
        msg = lex.error_leader() + "Empty token"
        raise RuntimeError, msg
    if tok[0] == "\"":
        val = tok.strip("\"")
        if not tok.endswith("\""):
            msg = lex.error_leader() + "String token not ending wiht '\"'"
            raise RuntimeError, msg
    elif tok == "(":
        val = parse_list(lex)
    else:  # should be a Hugin expression (currently a number or NoisyOR)
        if tok == "NoisyOR":
            val = ["NoisyOR"]
            tok2 = lex.get_token()
            check_token("(", tok2, lex)
            while tok2 != ")":
                name = lex.get_token()
                comma = lex.get_token()
                check_token(",", comma, lex)
                tok2 = lex.get_token()
                try:
                    prob = float(tok2)
                except ValueError:
                    msg = lex.error_leader() + "Wrong probability value '" + tok2 + "'"
                    raise RuntimeError, msg
                tok2 = lex.get_token()
                if tok2 != ")" and tok2 != ",":
                    msg = lex.error_leader() + "Expected ')' or ',', got '" + tok2 + "'"
                    raise RuntimeError, msg
                val.append((name, prob))
        else: # should be a number
            try:
                val = float(tok)
            except ValueError:
                msg = lex.error_leader() + "Wrong number string '" + tok + "'"
                raise RuntimeError, msg
    return val

def parse_node(node, lex):
    match_token("{", lex)
    tok = lex.get_token()
    while tok != "}" and tok != "":
	param = tok
	match_token("=", lex)
	value = parse_value(lex)
	node.params[param] = value
        match_token(";", lex)
	tok = lex.get_token()
    check_token("}", tok, lex)


def parse_attrs(lex):
    """Parse attributes of a potential."""
    match_token("(", lex)
    target = lex.get_token()
    tok = lex.get_token()
    conditioned_on = []
    if tok == "|":
        tok = lex.get_token()
        while tok != ")" and tok != "":
            conditioned_on.append(tok)
            tok = lex.get_token()
    check_token(")", tok, lex)
    return (target, conditioned_on)

class HuginNode(object):
    def __init__(self, _class, name, node_type = "discrete"):
        self._class = _class
        self.node_type = node_type
        self.name = name
        self.params = {}
        self.target = None
        self.conditioned_on = None
    def __str__(self):
        ret = ""
        ret += self._class + " " + self.node_type + " " + self.name
        if self.target != None:
            ret += "(%s | %s)" % (self.target, str(self.conditioned_on))
        ret += "\n" + str(self.params)
        return ret
    def __repr__(self):
        return self.__str__()

def parse_net(net, lex):
    tok = lex.get_token()
    while tok != "":
        nodetype = "discrete"
        if tok in ["discrete", "continuous"]:
            nodetype = tok
            tok = lex.get_token()
        nodeclass = tok
        if nodeclass == "potential":
            attrs = parse_attrs(lex)
        tok = lex.get_token()
        if tok == "{":
            nodename = ""
            lex.push_token(tok)
        else:
            nodename = tok
        node = HuginNode(nodeclass, nodename, nodetype)
        if nodeclass == "potential":
            node.target = attrs[0]
            node.conditioned_on = attrs[1]
        parse_node(node, lex)
        net.append(node)
	tok = lex.get_token()    


def read_Hugin_file(file_name):
    """Return a BayesNet read from a Hugin .net file.

    Files ending with .gz are treated as gzip compressed."""
    # parse the file
    try:
        if file_name.endswith(".gz"):
            import gzip
            file = gzip.GzipFile(file_name)
        else:
            file = open(file_name, "rU")
    except IOError, ie:
        raise BNReadError, "Error reading bayesian network: " + ie.strerror + ":" + ie.filename
    lex = shlex.shlex(file, file_name)
    lex.commenters = "%"
    lex.wordchars += "-+."
    lex.quotes = "\""
    net = []
    parse_net(net, lex)

    ### build the network
    # get nodes
    attrs = []
    domains = {}
    ids_2_names = {} # map node id's to node labels
    ids_2_numbers = {}
    for n in net:
        if n._class == "node": 
            if n.node_type != "discrete":
                raise RuntimeError, "Only discrete nodes are supported"
	    name = n.params.get('label', n.name)
	    if ids_2_names.has_key(n.name):
		raise RuntimeError, "Repeated definition for node " + n.name
	    ids_2_names[n.name] = name
            ids_2_numbers[n.name] = len(attrs)
            attrs.append(Attr(name, "CATEG", n.params['states']))
            domains[n.name] = n.params['states']
    bn = BayesNet(file_name, attrs)
    # get potentials
    for p in net:
        if p._class == "potential":
            shape = []
            for a in p.conditioned_on + [p.target]:
                shape.append(len(domains[a]))
            if   p.params.has_key('data'):
                distr = numpy.array(p.params['data'])
            elif p.params.has_key('model_data'):
                nor = NoisyOR(p.conditioned_on, shape[:-1])
                for vname, prb in p.params['model_data'][0][1:]:
                    nor.add_variable(vname, prb)
                distr = nor.get_table()
            else:
                raise RuntimeError, "No distribution for node " + str(ids_2_names[p.target])
            distr.shape = shape
	    conditioned_on = [ids_2_numbers[c] for c in p.conditioned_on]
            
            bn[ids_2_numbers[p.target]].set_parents_distr(conditioned_on, distr)
    return bn


def write_Hugin_file(bn, f):
    """Write a Bayesian network bn to a file f in Hugin format."""
    print >>f, """net
{
    node_size = (80 40);
    HR_Grid_X = "10";
    HR_Grid_Y = "10";
    HR_Grid_GridSnap = "1";
    HR_Grid_GridShow = "0";
    HR_Font_Name = "Arial";
    HR_Font_Size = "-12";
    HR_Font_Weight = "400";
    HR_Font_Italic = "0";
    HR_Propagate_Auto = "1";
}"""
    for i, n in enumerate(bn):
        print >>f, "\nnode V_" + str(i)
        print >>f, "{"
        print >>f, "  states = (" + " ".join(["\""+x+"\"" for x in n.domain]) + ");"
        print >>f, "  label = \"" + n.name + "\";"
        print >>f, "  position = (0 0);"
        print >>f, "}"
    for i, n in enumerate(bn):
        if len(n.parents) > 0:
            cond = "| " + " ".join(["V_"+str(x) for x in n.parents])
        else:
            cond = ""
        print >>f, "\npotential ( V_" + str(i) + cond + ")"
        print >>f, "{"
        print >>f, "  data = (" + " ".join([str(x) for x in n.distr.flat]) + ");"
        print >>f, "}"


def main():
    #bn = read_Hugin_file("oil.net")
    bn = read_Hugin_file("../BayesPrune/data/ksl_discr.net")
    #bn = read_Hugin_file("ksl.net")
    #bn = read_Hugin_file("/home/szymon/dmining/Python/BayesPrune/data/soybean.net")
    #bn = read_Hugin_file("/home/szymon/dmining/data/Borreliosis/borrelia200603.net")
    print bn
    #raise BNReadError, "txt"
    f = file("/tmp/bn", "w")
    write_Hugin_file(bn, f)

if __name__ == "__main__":
    main()
